import datetime

from ckan.common import _
from ckan.lib.plugins import DefaultTranslation
import ckan.plugins as plugins
import ckan.plugins.toolkit as toolkit

from ckanext.ark import cli, views
from ckanext.ark.lib.helpers import build_erc_metadata, get_ark_url
from ckanext.ark.model.crud import ARKQuery


class ArkPlugin(plugins.SingletonPlugin, DefaultTranslation):
    plugins.implements(plugins.IConfigurer)
    plugins.implements(plugins.IBlueprint, inherit=True)
    plugins.implements(plugins.IClick)
    plugins.implements(plugins.IPackageController, inherit=True)
    plugins.implements(plugins.ITranslation)

    # IConfigurer

    def update_config(self, config_):
        toolkit.add_template_directory(config_, 'templates')

    # IBlueprint

    def get_blueprint(self):
        return [views.blueprints]

    # IClick

    def get_commands(self):
        return cli.get_commands()

    # IPackageController

    def after_update(self, context, pkg_dict):
        '''Dataset has been created/updated. Check status of the dataset to
        determine if we should mint ARK.
        '''
        # Is this active and public? If so we need to make sure we have
        # an active ARK
        if (pkg_dict.get('state', 'active') == 'active' and
                not pkg_dict.get('private', False)):
            package_id = pkg_dict['id']

            # Remove user-defined update schemas first (if needed)
            context.pop('schema', None)

            # Load the package_show version of the dict
            pkg_show_dict = toolkit.get_action('package_show')(context, {
                'id': package_id
            })

            erc_record = build_erc_metadata(pkg_show_dict)

            if not erc_record:
                return pkg_dict

            # Load or create the ARK (package may not have a ARK if
            # extension was loaded after package creation)
            ark = ARKQuery.read_package(package_id, create_if_none=True)

            if ark.last_modified is None:
                toolkit.h.flash_success(_('ARK identifier created'))

            ARKQuery.update_ark(identifier=ark.identifier,
                                last_modified=datetime.datetime.utcnow(),
                                **erc_record
                                )

        return pkg_dict

    def after_show(self, context, pkg_dict):
        '''Add the ARK details to the pkg_dict so it can be displayed.
        '''
        ark = ARKQuery.read_package(pkg_dict['id'])
        if ark:
            pkg_dict.update({
                'ark': f'ark:{ark.identifier}',
                'erc_who': ark.who,
                'erc_what': ark.what,
                'erc_when': ark.when,
                'erc_where': get_ark_url(ark.identifier)
            })



from ckan.model import Session

from ckanext.ark.model.ark import ARK, ark_table


class ARKQuery:
    '''Query methods for ARK table.
    '''
    m = ARK
    cols = [c.name for c in ark_table.c]

    @classmethod
    def create(cls, identifier, package_id, who='', what='', when=''):
        '''Create a new record in the ARK table.

        :param identifier: the compact ARK without the 'ark:' part
        :type identifier: string
        :param package_id: the id of the package this ARK represents
        :type package_id: string

        :returns: the newly created record object
        '''
        new_record = ARK(identifier=identifier,
                         package_id=package_id,
                         who=who,
                         what=what,
                         when=when)
        Session.add(new_record)
        Session.commit()
        return new_record

    @classmethod
    def read_ark(cls, identifier):
        '''Retrieve a record with a given ARK.

        :param identifier: the compact ARK without the 'ark:' part
        :type identifier: string

        :returns: the record object
        '''
        return Session.query(ARK).get(identifier)

    @classmethod
    def read_package(cls, package_id, create_if_none=False):
        '''Retrieve a record associated with a given package.

        :param package_id: the id of the package
        :type package_id: string
        :param create_if_none: generate a new ARK and add a record if no record
                               is found for the given package
                               (optional, default: False)
        :type create_if_none: bool

        :returns: the record object
        '''
        from ckanext.ark.lib.minter import Minter
        record = (Session.query(ARK).filter(ARK.package_id == package_id)
                  .first())
        if record is None and create_if_none:
            minter = Minter()
            new_ark = minter.mint_ark()
            record = cls.create(identifier=new_ark, package_id=package_id)
        return record

    @classmethod
    def update_ark(cls, identifier, **kwargs):
        '''Update the published fields of a record with a given ARK.

        :param identifier: the compact ARK without the 'ark:' part
        :type identifier: string
        :param kwargs: the values to be updated
        :type kwargs: dictionary

        :returns: the updated record object
        '''
        update_dict = {k: v for k, v in kwargs.items() if k in cls.cols}
        (Session.query(ARK).filter(ARK.identifier == identifier)
         .update(update_dict))
        Session.commit()
        return cls.read_ark(identifier)

    @classmethod
    def delete_ark(cls, identifier):
        '''Delete the record with a given ARK.

        :param identifier: the compact ARK without the 'ark:' part
        :type identifier: string

        :returns: True if a record was deleted, False if not
        :rtype: bool
        '''
        to_delete = cls.read_ark(identifier)
        if to_delete is not None:
            Session.delete(to_delete)
            Session.commit()
            return True
        else:
            return False



import json

from ckan.plugins import toolkit


def _get_nma_url():
    '''Return the NMA URL. Try and use ckanext.ark.nma_url but if that's not set
    use ckan.site_url.

    :returns: the NMA URL
    :rtype: string
    '''
    nma_url = toolkit.config.get('ckanext.ark.nma_url',
                                 toolkit.config.get('ckan.site_url'))
    return nma_url.rstrip('/')


def _get_mapping_fields_dict():
    '''Return a mapping from CKAN fields to ERC metadata fields.

    :returns: a mapping from CKAN fields to ERC metadata fields
    :rtype: dictionary
    '''
    ckan_to_erc = {'who': 'author', 'what': 'title'}
    erc_fields = ['who', 'what', 'when']
    mappings = toolkit.config.get('ckanext.ark.erc_mappings')
    if mappings:
        if isinstance(mappings, str):
            ckan_to_erc.update(json.loads(mappings))
        elif isinstance(mappings, dict):
            ckan_to_erc.update(mappings)
    return {k: v for k, v in ckan_to_erc.items() if k in erc_fields}


def _is_allow_missing_erc():
    '''Return True if missing ERC metadata is allowed for assigning ARKs.

    :returns: True if missing ERC metadata is allowed, False if not
    :rtype: bool
    '''
    return toolkit.asbool(
        toolkit.config.get('ckanext.ark.allow_missing_erc', True))


def build_erc_metadata(pkg_show_dict):
    '''Return the ERC metadata.

    :param pkg_show_dict: the package dictionary
    :type pkg_show_dict: dictionary

    :returns: the ERC metadata
    :rtype: dictionary
    '''
    erc_record = {}
    allow_missing_erc = _is_allow_missing_erc()

    for erc_field, ckan_field in _get_mapping_fields_dict().items():
        if erc_field == 'when' and isinstance(ckan_field, list):
            if len(ckan_field) != 2:
                return
            when_from = pkg_show_dict.get(ckan_field[0], '')
            when_to = pkg_show_dict.get(ckan_field[1], '')
            if not when_from and not when_to and not allow_missing_erc:
                return
            when_from = when_from.replace('-', '')
            when_to = when_to.replace('-', '')
            if when_from and when_to:
                when_from += '-'
            erc_record[erc_field] = f'{when_from}{when_to}'
        elif ckan_field in pkg_show_dict.keys():
            erc_record[erc_field] = pkg_show_dict[ckan_field]
        elif not allow_missing_erc:
            return

    return erc_record


def get_ark_url(identifier):
    '''Return the ARK URL.

    :param identifier: the compact ARK without the 'ark:' part
    :type identifier: string

    :returns: the ARK URL
    :rtype: string
    '''
    return f'{_get_nma_url()}/ark:{identifier}'


def get_erc_support():
    '''Return the ERC support metadata.

    :returns: the ERC support metadata
    :rtype: dictionary
    '''
    naan = toolkit.config.get('ckanext.ark.naan')
    erc_support = {
        'who': toolkit.config.get('ckanext.ark.erc_support.who', ''),
        'what': toolkit.config.get('ckanext.ark.erc_support.what', ''),
        'when': toolkit.config.get('ckanext.ark.erc_support.when', ''),
        'where': f'{_get_nma_url()}/ark:{naan}'
    }
    return erc_support


def get_erc_support_commitment():
    '''Return the persistence statement from the NMA.

    :returns: the persistence statement from the NMA
    :rtype: string
    '''
    return toolkit.config.get('ckanext.ark.erc_support.commitment', '')



from noid import pynoid as noid
from ckan.plugins import toolkit

from ckanext.ark.model.crud import ARKQuery


class Minter:
    '''A minter for ARK.
    '''
    def __init__(self):
        self.naan = toolkit.config.get('ckanext.ark.naan')
        self.shoulder = toolkit.config.get('ckanext.ark.shoulder', '')
        self.template = toolkit.config.get('ckanext.ark.template')

        if self.naan is None:
            raise TypeError('You must set the ckanext.ark.naan config value')

    def mint_ark(self):
        '''Create a new ARK which isn't currently in use.
        '''
        attempts = 5

        while attempts > 0:
            if self.template:
                blade = noid.mint(template=self.template)
            else:
                blade = noid.mint()
            ark = f'{self.naan}/{self.shoulder}{blade}'

            if ARKQuery.read_ark(ark) is None:
                return ark
            attempts -= 1
        raise Exception('Failed to create an ARK identifier')



import datetime

from ckan.model import core, meta
from ckan.model.domain_object import DomainObject
from sqlalchemy import Column, Table, types


ark_table = Table(
    'ark',
    meta.metadata,
    Column('identifier', types.UnicodeText, primary_key=True),
    Column('package_id', types.UnicodeText, nullable=False, unique=True),
    Column('who', types.UnicodeText),
    Column('what', types.UnicodeText),
    Column('when', types.UnicodeText),
    Column('created', types.DateTime, default=datetime.datetime.utcnow),
    Column('last_modified', types.DateTime),
    Column('state', types.UnicodeText, default=core.State.ACTIVE),
)


class ARK(DomainObject):
    '''ARK Object'''
    pass


meta.mapper(ARK, ark_table)



# -*- coding: utf-8 -*-
# Always prefer setuptools over distutils
from setuptools import setup, find_namespace_packages
from codecs import open  # To use a consistent encoding
from os import path

here = path.abspath(path.dirname(__file__))

# Get the long description from the relevant file
with open(path.join(here, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

setup(
    name='ckanext-ark',
    version='1.0.1',
    description='A CKAN extension for assigning Archival Resource Key '
                '(ARK) identifiers to datasets.',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/depositar/ckanext-ark',
    author='Cheng-Jen Lee',
    author_email='u103133.u103135@gmail.com',
    license='MIT',
    classifiers=[
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
    ],
    keywords='CKAN ARK',
    packages=find_namespace_packages(exclude=['ckanext.ark.tests']),
    namespace_packages=['ckanext'],
    install_requires=[
        'noid-mint'
    ],
    include_package_data=True,
    entry_points='''
        [ckan.plugins]
        ark=ckanext.ark.plugin:ArkPlugin

        [babel.extractors]
        ckan = ckan.lib.extract:extract_ckan
    ''',
    message_extractors={
        'ckanext': [
            ('**.py', 'python', None),
            ('**.js', 'javascript', None),
            ('**/templates/**.html', 'ckan', None),
        ],
    }
)



from flask import Blueprint, make_response, request
from ckan.common import _
from ckan.lib import base
from ckan.plugins import toolkit

from ckanext.ark.lib.helpers import get_ark_url, get_erc_support, \
    get_erc_support_commitment
from ckanext.ark.model.crud import ARKQuery

blueprints = Blueprint('ark', __name__)


@blueprints.route('/ark:/<path:path>/')
@blueprints.route('/ark:<path:path>/')
def read(path):
    # Show NAA metadata
    if path == toolkit.config.get('ckanext.ark.naan'):
        response = make_response(get_erc_support_commitment())
        response.headers['Content-type'] = 'text/plain; charset=UTF-8'
        return response
    ark = ARKQuery.read_ark(path)
    if not ark:
        return base.abort(404, _('ARK not found'))
    # Show ERC metadata
    if 'info' in request.args or request.environ['REQUEST_URI'][-2:] == '/?':
        response = {
            'erc': {
                'who': ark.who,
                'what': ark.what,
                'when': ark.when,
                'where': get_ark_url(ark.identifier)
            },
            'erc-support': get_erc_support()
        }
        response = make_response(response)
        response.headers['Content-type'] = 'application/json; charset=UTF-8'
        return response
    else:
        try:
            toolkit.get_action('package_show')({}, {
                'id': ark.package_id
            })
            return toolkit.redirect_to('dataset.read',
                                       id=ark.package_id)
        except (toolkit.ObjectNotFound, toolkit.NotAuthorized):
            # Show defunct page
            return toolkit.render('ark/snippets/defunct.html',
                                  {'ark': ark.identifier})



import click
from ckan import model
from ckan.plugins import toolkit

from ckanext.ark.lib.helpers import build_erc_metadata
from ckanext.ark.model import ark as ark_model
from ckanext.ark.model.crud import ARKQuery
from ckanext.ark.model.ark import ARK


def get_commands():
    return [ark]


@click.group()
def ark():
    '''Ark commands.
    '''
    pass


@ark.command(name='initdb')
def init_db():
    if ark_model.ark_table.exists():
        click.secho('ARK table already exists', fg='green')
    else:
        ark_model.ark_table.create()
        click.secho('ARK table created', fg='green')


@ark.command(name='update-ark',
             short_help='Give ARK identifiers to existed datasets')
def update_ark():
    '''Mint ARKs for existed packages.
    '''
    context = {'model': model, 'ignore_auth': True, 'use_cache': False}

    package_ids = [r[0] for r in model.Session.query(model.Package.id).
                   filter(model.Package.type == 'dataset').all()]

    pkg_ids_with_arks = [r[0] for r in
                         model.Session.query(ARK.package_id).all()]

    pkgs_to_update = set(package_ids) - set(pkg_ids_with_arks)

    if len(pkgs_to_update) == 0:
        click.secho('No datasets found to update', fg='green')
        return

    for package_id in pkgs_to_update:
        pkg_dict = toolkit.get_action('package_show')(context, {
            'id': package_id
        })
        title = pkg_dict.get('title', package_id)

        if (pkg_dict.get('state', 'active') != 'active' or
                pkg_dict.get('private', False)):
            click.secho(f'"{title}" is inactive or private; ignoring',
                        fg='yellow')
            continue

        erc_record = build_erc_metadata(pkg_dict)

        if not erc_record:
            click.secho(
                f'"{title}" does not meet the erc requirements; ignoring',
                fg='yellow')
            continue

        ark = ARKQuery.read_package(package_id, create_if_none=True)

        click.secho(
            f'Updated "{title}" with ARK identifier "{ark.identifier}"',
            fg='green')


@ark.command(name='delete-ark',
             short_help='Delete ARK identifier for existed dataset')
@click.argument('name')
def delete_ark(name):
    '''Delete ARK for existed package. Accept package's id (name) and
    ARK identifier (with and without ark:).
    '''
    identifier = name.replace('ark:', '')
    if ARKQuery.delete_ark(identifier):
        click.secho(f'Deleted ARK {name} from the database')
        return
    dataset = model.Package.get(name)
    if dataset:
        ark = ARKQuery.read_package(dataset.id, create_if_none=True)
        if ARKQuery.delete_ark(ark.identifier):
            click.secho(f'Deleted ARK ark:{ark.identifier} from the database')
            return
    click.secho('Nothing to delete', fg='green')
